@@ -1,7 +1,10 @@
 /* -*- Mode: C; tab-width: 4; c-basic-offset: 4; indent-tabs-mode: nil -*- */
-/* $Id: memcached.h 570 2007-06-20 01:21:45Z plindner $ */
+/* $Id: memcached.h 651 2007-11-21 19:59:12Z dormando $ */
 
+#ifdef HAVE_CONFIG_H
 #include "config.h"
+#endif
+
 #include <sys/types.h>
 #include <sys/socket.h>
 #include <sys/time.h>
@@ -13,17 +16,23 @@
 #define UDP_MAX_PAYLOAD_SIZE 1400
 #define UDP_HEADER_SIZE 8
 #define MAX_SENDBUF_SIZE (256 * 1024 * 1024)
+/* I'm told the max legnth of a 64-bit num converted to string is 20 bytes.
+ * Plus a few for spaces, \r\n, \0 */
+#define SUFFIX_SIZE 24
 
-/* Initial size of list of items being returned by "get". */
+/** Initial size of list of items being returned by "get". */
 #define ITEM_LIST_INITIAL 200
 
-/* Initial size of the sendmsg() scatter/gather array. */
+/** Initial size of list of CAS suffixes appended to "gets" lines. */
+#define SUFFIX_LIST_INITIAL 20
+
+/** Initial size of the sendmsg() scatter/gather array. */
 #define IOV_LIST_INITIAL 400
 
-/* Initial number of sendmsg() argument structures to allocate. */
+/** Initial number of sendmsg() argument structures to allocate. */
 #define MSG_LIST_INITIAL 10
 
-/* High water marks for buffer shrinking */
+/** High water marks for buffer shrinking */
 #define READ_BUFFER_HIGHWAT 8192
 #define ITEM_LIST_HIGHWAT 400
 #define IOV_LIST_HIGHWAT 600
@@ -47,7 +56,7 @@
 # include <unistd.h>
 #endif
 
-/* Time relative to server start. Smaller than time_t on 64-bit systems. */
+/** Time relative to server start. Smaller than time_t on 64-bit systems. */
 typedef unsigned int rel_time_t;
 
 struct stats {
@@ -80,6 +89,7 @@ struct settings {
     bool managed;          /* if 1, a tracker manages virtual buckets */
     int evict_to_free;
     char *socketpath;   /* path to unix socket if using local socket */
+    int access;  /* access mask (a la chmod) for unix domain socket */
     double factor;          /* chunk size growth factor */
     int chunk_size;
     int num_threads;        /* number of libevent threads to run */
@@ -108,6 +118,7 @@ typedef struct _stritem {
     uint8_t         it_flags;   /* ITEM_* above */
     uint8_t         slabs_clsid;/* which slab class we're in */
     uint8_t         nkey;       /* key length, w/terminating null and padding */
+    uint64_t        cas_id;     /* the CAS identifier */
     void * end[];
     /* then null-terminated key */
     /* then " flags length\r\n" (no terminating null) */
@@ -122,44 +133,47 @@ typedef struct _stritem {
 #define ITEM_ntotal(item) (sizeof(struct _stritem) + (item)->nkey + 1 + (item)->nsuffix + (item)->nbytes)
 
 enum conn_states {
-    conn_listening,  /* the socket which listens for connections */
-    conn_read,       /* reading in a command line */
-    conn_write,      /* writing out a simple response */
-    conn_nread,      /* reading in a fixed number of bytes */
-    conn_swallow,    /* swallowing unnecessary bytes w/o storing */
-    conn_closing,    /* closing this connection */
-    conn_mwrite      /* writing out many items sequentially */
+    conn_listening,  /** the socket which listens for connections */
+    conn_read,       /** reading in a command line */
+    conn_write,      /** writing out a simple response */
+    conn_nread,      /** reading in a fixed number of bytes */
+    conn_swallow,    /** swallowing unnecessary bytes w/o storing */
+    conn_closing,    /** closing this connection */
+    conn_mwrite,     /** writing out many items sequentially */
 };
 
 #define NREAD_ADD 1
 #define NREAD_SET 2
 #define NREAD_REPLACE 3
+#define NREAD_APPEND 4
+#define NREAD_PREPEND 5
+#define NREAD_CAS 6
 
 typedef struct {
     int    sfd;
     int    state;
     struct event event;
     short  ev_flags;
-    short  which;   /* which events were just triggered */
+    short  which;   /** which events were just triggered */
 
-    char   *rbuf;   /* buffer to read commands into */
-    char   *rcurr;  /* but if we parsed some already, this is where we stopped */
-    int    rsize;   /* total allocated size of rbuf */
-    int    rbytes;  /* how much data, starting from rcur, do we have unparsed */
+    char   *rbuf;   /** buffer to read commands into */
+    char   *rcurr;  /** but if we parsed some already, this is where we stopped */
+    int    rsize;   /** total allocated size of rbuf */
+    int    rbytes;  /** how much data, starting from rcur, do we have unparsed */
 
     char   *wbuf;
     char   *wcurr;
     int    wsize;
     int    wbytes;
-    int    write_and_go; /* which state to go into after finishing current write */
-    void   *write_and_free; /* free this memory after finishing writing */
+    int    write_and_go; /** which state to go into after finishing current write */
+    void   *write_and_free; /** free this memory after finishing writing */
 
-    char   *ritem;  /* when we read in an item's value, it goes here */
+    char   *ritem;  /** when we read in an item's value, it goes here */
     int    rlbytes;
 
     /* data for the nread state */
 
-    /*
+    /**
      * item is used to hold an item structure created after reading the command
      * line of set/add/replace commands, but before we finished reading the actual
      * data. The data is read into ITEM_data(item) to avoid extra copying.
@@ -187,6 +201,11 @@ typedef struct {
     item   **icurr;
     int    ileft;
 
+    char   **suffixlist;
+    int    suffixsize;
+    char   **suffixcurr;
+    int    suffixleft;
+
     /* data for UDP clients */
     bool   udp;       /* is this is a UDP "connection" */
     int    request_id; /* Incoming UDP request ID, if this is a UDP "connection" */
@@ -216,10 +235,12 @@ extern volatile rel_time_t current_time;
  */
 
 conn *do_conn_from_freelist();
-int do_conn_add_to_freelist(conn *c);
+bool do_conn_add_to_freelist(conn *c);
+char *do_suffix_from_freelist();
+bool do_suffix_add_to_freelist(char *s);
 char *do_defer_delete(item *item, time_t exptime);
 void do_run_deferred_deletes(void);
-char *do_add_delta(item *item, int incr, const unsigned int delta, char *buf);
+char *do_add_delta(item *item, const bool incr, const int64_t delta, char *buf);
 int do_store_item(item *item, int comm);
 conn *conn_new(const int sfd, const int init_state, const int event_flags, const int read_buffer_size, const bool is_udp, struct event_base *base);
 
@@ -250,18 +271,23 @@ int  dispatch_event_add(int thread, conn *c);
 void dispatch_conn_new(int sfd, int init_state, int event_flags, int read_buffer_size, int is_udp);
 
 /* Lock wrappers for cache functions that are called from main loop. */
-char *mt_add_delta(item *item, const int incr, const unsigned int delta, char *buf);
+char *mt_add_delta(item *item, const int incr, const int64_t delta, char *buf);
 void mt_assoc_move_next_bucket(void);
 conn *mt_conn_from_freelist(void);
-int   mt_conn_add_to_freelist(conn *c);
+bool  mt_conn_add_to_freelist(conn *c);
+char *mt_suffix_from_freelist(void);
+bool  mt_suffix_add_to_freelist(char *s);
 char *mt_defer_delete(item *it, time_t exptime);
 int   mt_is_listen_thread(void);
 item *mt_item_alloc(char *key, size_t nkey, int flags, rel_time_t exptime, int nbytes);
+char *mt_item_cachedump(const unsigned int slabs_clsid, const unsigned int limit, unsigned int *bytes);
 void  mt_item_flush_expired(void);
 item *mt_item_get_notedeleted(const char *key, const size_t nkey, bool *delete_locked);
 int   mt_item_link(item *it);
 void  mt_item_remove(item *it);
 int   mt_item_replace(item *it, item *new_it);
+char *mt_item_stats(int *bytes);
+char *mt_item_stats_sizes(int *bytes);
 void  mt_item_unlink(item *it);
 void  mt_item_update(item *it);
 void  mt_run_deferred_deletes(void);
@@ -278,14 +304,19 @@ int   mt_store_item(item *item, int comm);
 # define assoc_move_next_bucket()    mt_assoc_move_next_bucket()
 # define conn_from_freelist()        mt_conn_from_freelist()
 # define conn_add_to_freelist(x)     mt_conn_add_to_freelist(x)
+# define suffix_from_freelist()      mt_suffix_from_freelist()
+# define suffix_add_to_freelist(x)   mt_suffix_add_to_freelist(x)
 # define defer_delete(x,y)           mt_defer_delete(x,y)
 # define is_listen_thread()          mt_is_listen_thread()
 # define item_alloc(x,y,z,a,b)       mt_item_alloc(x,y,z,a,b)
+# define item_cachedump(x,y,z)       mt_item_cachedump(x,y,z)
 # define item_flush_expired()        mt_item_flush_expired()
 # define item_get_notedeleted(x,y,z) mt_item_get_notedeleted(x,y,z)
 # define item_link(x)                mt_item_link(x)
 # define item_remove(x)              mt_item_remove(x)
 # define item_replace(x,y)           mt_item_replace(x,y)
+# define item_stats(x)               mt_item_stats(x)
+# define item_stats_sizes(x)         mt_item_stats_sizes(x)
 # define item_update(x)              mt_item_update(x)
 # define item_unlink(x)              mt_item_unlink(x)
 # define run_deferred_deletes()      mt_run_deferred_deletes()
@@ -304,16 +335,21 @@ int   mt_store_item(item *item, int comm);
 # define assoc_move_next_bucket()    do_assoc_move_next_bucket()
 # define conn_from_freelist()        do_conn_from_freelist()
 # define conn_add_to_freelist(x)     do_conn_add_to_freelist(x)
+# define suffix_from_freelist()      do_suffix_from_freelist()
+# define suffix_add_to_freelist(x)   do_suffix_add_to_freelist(x)
 # define defer_delete(x,y)           do_defer_delete(x,y)
 # define dispatch_conn_new(x,y,z,a,b) conn_new(x,y,z,a,b,main_base)
 # define dispatch_event_add(t,c)     event_add(&(c)->event, 0)
 # define is_listen_thread()          1
 # define item_alloc(x,y,z,a,b)       do_item_alloc(x,y,z,a,b)
+# define item_cachedump(x,y,z)       do_item_cachedump(x,y,z)
 # define item_flush_expired()        do_item_flush_expired()
 # define item_get_notedeleted(x,y,z) do_item_get_notedeleted(x,y,z)
 # define item_link(x)                do_item_link(x)
 # define item_remove(x)              do_item_remove(x)
 # define item_replace(x,y)           do_item_replace(x,y)
+# define item_stats(x)               do_item_stats(x)
+# define item_stats_sizes(x)         do_item_stats_sizes(x)
 # define item_unlink(x)              do_item_unlink(x)
 # define item_update(x)              do_item_update(x)
 # define run_deferred_deletes()      do_run_deferred_deletes()
